/*
 * List File Info.
 *
 * Copyright (c) 2003 - 2008 Greg Snyder.
 *
 * Home page:  http://www.awokengazebo.com/software/lfi/
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */


/*
 *
 * This program generates a list of file info lines.  These info lines
 * capture the interesting meta information about a file, as well as an MD5
 * of the file's data (or the link target name in the case of a symlink).
 *
 * This is useful as part of making a backup, or later verifying that
 * nothing about a file or group of files has changed from a previous backup.
 *
 * An example line is:  (wrapped for readability)
 *
 *     lfiv1 010 0644 1 102 100 - 419 1045176336 944704630 1045176336 
 *              md5 59da6e284a7e15b8b3a3afa659862958 /usr/home/gregs/.login
 *
 * The fields are:
 *     line id
 *     file type (octal)  (04=dir, 010=file, 012=symlink)
 *     file mode (octal) 
 *     link count
 *     uid
 *     gid
 *     device number (only for block or character special files, - otherwise)
 *     size (in bytes)
 *     access time (last time file contents changed)
 *     modification time (last time file contents changed)
 *     change time (last time file meta info (mode, owner, etc))
 *     hash name (Or - when there is no hash)
 *     hash (for symlinks: hash of link target, '-' for directories)
 *     file name  (with special character escaped)
 *
 * The fields from the stat structure which are missing are:
 *     dev: raw device number
 *     ino: inode number
 *
 * These fields are not listed because they generally would not aid in
 *     file backup/restore or comparing files between systems.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>


static const char rcsid[] = "$Id: lfi.c,v 1.32 2008/07/23 17:06:15 gregs Exp $";

extern char **environ;

#ifndef min
#define min(x,y) ( (x)<(y)?(x):(y) )
#endif

#define LFI_MAX_HASH_PROG_ARGS 16
#define LFI_MAX_HASH_ARG_LENGTH 256
#define LFI_PATH_MAX (min(PATH_MAX,256))

/* Constants for identifying the different hash choices */
typedef struct lfi_hash_record {
    int lhr_hash_id;               /* A unique number for this record */
    const char lhr_hash_descr[64]; /* Internal description */
    const char lhr_hash_name[32];  /* String put into lfi output line */
    unsigned int lhr_hash_length;  /* Length in bytes of the output'd hash */

    const char lhr_prog_binary_path[LFI_PATH_MAX];

    /* "%p" will be substituted for the full file path */
    const char *lhr_prog_args_filename[LFI_MAX_HASH_PROG_ARGS]; 

    /* stdin is hashed instead of a file */
    const char *lhr_prog_args_stdin[LFI_MAX_HASH_PROG_ARGS];    

    /*
     * 1: Linux 'md5sum': bbad415e1c29bbedd9b383537100412c  /foo
     * 2: FreeBSD 'md5' : MD5 (/foo) = bbad415e1c29bbedd9b383537100412c
     * 3: openssl dgst  : MD5(/foo)= bbad415e1c29bbedd9b383537100412c
     * 
     */

    /* for filenames, num bytes (besides pathname and hash) in raw line */
    int lhr_hash_parse_filename_extra_min;   

    /* If the raw output also has the pathname */
    unsigned char lhr_hash_parse_filename_has_path;  

    /* use 'hash_length' bytes starting from here.
     * '0' is first character, and '-10' would be ten
     * characters from the end.
     */
    int lhr_hash_parse_filename_use_start;   

    /* for stdin, number of bytes (besides pathname and hash) in raw line */
    int lhr_hash_parse_stdin_extra_min;      

    /* If the raw output also has the pathname */
    unsigned char lhr_hash_parse_stdin_has_path;  

    /* use 'hash_length' bytes starting from here.
     * '0' is first character, and '-10' would be ten
     * characters from the end.
     */
    int lhr_hash_parse_stdin_use_start;   

    /* In raw output, how stdin is shown, likely "" or "-" */
    const char lhr_hash_parse_stdin_filename[LFI_PATH_MAX]; 
} lfi_hash_record;


const lfi_hash_record lfi_hash_records[] = {
#ifdef __FreeBSD__
    {
        10,
        "FreeBSD md5",
        "md5",
        32,
        "/sbin/md5",
        {"md5", "--", "%p", NULL},
        {"md5", NULL},
        9, 1, -32,
        0, 0, 0, ""
    },
    {
        20,
        "FreeBSD sha1",
        "sha1",
        40,
        "/sbin/sha1",
        {"sha1", "--", "%p", NULL},
        {"sha1", NULL},
        10, 1, -40,
        0, 0, 0, ""
    },
    {
        30,
        "FreeBSD sha256",
        "sha256",
        64,
        "/sbin/sha256",
        {"sha256", "--", "%p", NULL},
        {"sha256", NULL},
        12, 1, -64,
        0, 0, 0, ""
    },
#endif
    {
        100,
        "GNU coreutils md5sum",
        "md5",
        32,
        "/usr/bin/md5sum",
        {"md5sum", "--", "%p", NULL},
        {"md5sum", NULL},
        2, 1, 0,
        2, 1, 0, "-"
    },
#if 0
    {
        110,
        "OpenSSL md5",
        "md5",
        32,
        "/usr/bin/openssl",
        {"openssl", "dgst", "-md5", "%p", NULL},
        {"openssl", "dgst", "-md5", NULL},
        7, 1, -32,
        0, 0, 0, ""
    },
#endif
    {
        120,
        "GNU coreutils sha1sum",
        "sha1",
        40,
        "/usr/bin/sha1sum",
        {"sha1sum", "--", "%p", NULL},
        {"sha1sum", NULL},
        2, 1, 0,
        2, 1, 0, "-"
    },
    {
        130,
        "GNU coreutils sha256sum",
        "sha256",
        64,
        "/usr/bin/sha256sum",
        {"sha256sum", "--", "%p", NULL},
        {"sha256sum", NULL},
        2, 1, 0,
        2, 1, 0, "-"
    },


};

/* Default hash name to use */
#define HASH_DEFAULT_NAME "md5"

#define FIELD_LFI_ID                          1
#define FIELD_LFI_TYPE                        2
#define FIELD_LFI_MODE                        3
#define FIELD_LFI_NLINK                       4
#define FIELD_LFI_UID                         5
#define FIELD_LFI_GID                         6
#define FIELD_LFI_RDEV                        7
#define FIELD_LFI_SIZE                        8
#define FIELD_LFI_ATIME                       9
#define FIELD_LFI_MTIME                       10
#define FIELD_LFI_CTIME                       11
#define FIELD_LFI_HASH_TYPE                   12
#define FIELD_LFI_HASH                        13
#define FIELD_LFI_FILENAME                    14

#define FIELD_LFI_LAST                        14
#define FIELD_LFI_MASK                        0xff
#define FIELD_LFI_FLAG(x)                     (1 << ((x) & FIELD_LFI_MASK))


#define FIELD_MATCH_TYPE_ALL              1
#define FIELD_MATCH_TYPE_FIFO             2
#define FIELD_MATCH_TYPE_CHAR_DEV         3
#define FIELD_MATCH_TYPE_DIR              4
#define FIELD_MATCH_TYPE_BLOCK_DEV        5
#define FIELD_MATCH_TYPE_FILE             6
#define FIELD_MATCH_TYPE_SYMLINK          7
#define FIELD_MATCH_TYPE_SOCKET           8
#define FIELD_MATCH_COUNT                 8
#define FIELD_MATCH_TYPE_MASK             0xff
#define FIELD_MATCH_FLAG(x)               (((x) & FIELD_MATCH_TYPE_MASK) << 20)

/* Set this to 0 if you do not want filenames escaped */
#define OPT_ESCAPE_FILENAME 1

/* ************************************************************ */

uid_t uid = -1;
short has_quiet = 0, has_verbose = 0;
 
#define MIN_STAT_BUFFER_LEN 64
#define S_FMT_SHIFT 12ul
#if ((S_IFMT  >> S_FMT_SHIFT) << S_FMT_SHIFT) != S_IFMT
#error fmt
#endif

static const char LFI_PREFIX[] = "lfi";
static const char LFI_ID[] = "lfiv1";

int escape_filename(const char *filename, int do_shell_esc, int ret_efn_len, 
                char *ret_efn);

int unescape_filename(const char *esc_filename, int ret_ufn_len, 
                      char *ret_ufn);

const char *hash_id_to_hash_name(int hash_id);

int hash_name_to_hash_id(const char *hash_name);

const lfi_hash_record *hash_record(int hash_id);

static int do_childhash(const char *filename, const char *in_buffer,
             int hash_id,
             char *ret_buf, int ret_buf_size,
             char *ret_error_buf, int ret_error_buf_size);

static int stat_mode_to_file_type(int stat_mode);

int stat_string_create(const char *filename, const char *link_filename, 
                   const struct stat *ss, 
                   int hash_id,
                   int exclude_mask,
                   int ret_str_len, char *ret_str);

int stat_string_extract_fields(const char *stat_str, 
                           char **ret_id, char **ret_type,
                           char **ret_mode, char **ret_nlink,
                           char **ret_uid, char **ret_gid,
                           char **ret_rdev, char **ret_size,
                           char **ret_atime, char **ret_mtime,
                           char **ret_ctime, char **ret_hash_type,
                           char **ret_hash, char **ret_raw_filename,

                           int *ret_hash_id,
                           char **ret_filename);

int stats_string_diff_mask(const char *line1, const char *line2);

/* 
 * The goal of this function is to make it easier to cut-n-paste filenames
 * from the output into commands.
 */
int
escape_filename(const char *filename, int do_shell_esc, int ret_efn_len, 
                char *ret_efn) 
{
    const unsigned char *csrc = (const unsigned char *) filename;
    char *cdst = ret_efn;
    int fnsl = 0xfffffff;

    if (!filename || !ret_efn || ret_efn_len < (fnsl = strlen(filename))) {
        return(1);
    }

    while (csrc && *csrc) {
        /* Too-conservative size check */
        if (cdst - ret_efn > ret_efn_len - 5) {
            *cdst = '\0';
            return(1);
        }

        if (*csrc < ' ' || *csrc > '~') {
            *cdst++ = '\\';
            *cdst++ = '0';
            *cdst++ = '0' + (*csrc >> 6);
            *cdst++ = '0' + ( (*csrc & 0070) >> 3);
            *cdst++ = '0' + (*csrc & 0007);
            csrc++;
        }
        else if (*csrc == ' ' || *csrc == '\\') {
            *cdst++ = '\\';
            *cdst++ = *csrc++;
        }
        else if (do_shell_esc && 
                 (*csrc == '?' || 
                  *csrc == '!' ||
                  *csrc == '"' ||
                  *csrc == '$' ||
                  *csrc == '&' ||
                  *csrc == '(' ||
                  *csrc == ')' ||
                  *csrc == '*' ||
                  *csrc == ';' ||
                  *csrc == '<' ||
                  *csrc == '>' ||
                  *csrc == '[' ||
                  *csrc == '^' ||
                  *csrc == '`' ||
                  *csrc == '\'' ||
                  *csrc == '{' ||
                  *csrc == '}' ||
                  *csrc == '|' ||
                  *csrc == '~')) {
            *cdst++ = '\\';
            *cdst++ = *csrc++;
        } else {
            *cdst++ = *csrc++;
        }
    }
    *cdst = '\0';

    return(0);
}


/* 
 * The goal of this function is to make it easier to cut-n-paste filenames
 * from the output into commands.
 */
int
unescape_filename(const char *esc_filename, int ret_ufn_len, char *ret_ufn) 
{
    const unsigned char *csrc = (const unsigned char *) esc_filename;
    char *cdst = ret_ufn;
    int fnsl = 0xfffffff;
    int octo;

    if (!esc_filename || !ret_ufn || 
        ret_ufn_len < (fnsl = strlen(esc_filename))) {
        return (1);
    }

    while (csrc && *csrc) {
        if (*csrc == '\\') {
            csrc++;

            /* See if this is a zero, meaning 3 octal digits follow */
            if (*csrc == '0') {
                csrc++;

                octo = *csrc++ - '0';
                octo = octo * 8 + *csrc++ - '0';
                octo = octo * 8 + *csrc++ - '0';

                *cdst++ = octo;
            }
            else {
                *cdst++ = *csrc++;
            }
        }
        else {
            *cdst++ = *csrc++;
        }
    }

    *cdst = '\0';

    return (0);
}


const char *
hash_id_to_hash_name(int hash_id)
{
    static int last_idx = -1;  /* a cache of our last lookup */
    int idx = -1;

    if (hash_id < 0) {
        return("*ERROR*");
    }

    if (last_idx != -1 && 
        lfi_hash_records[last_idx].lhr_hash_id == hash_id) {
        return(lfi_hash_records[last_idx].lhr_hash_name);
    }

    for (idx = 0; 
         idx < (int) (sizeof(lfi_hash_records)/sizeof(lfi_hash_records[0]));
         idx++) {

        if (lfi_hash_records[idx].lhr_hash_id == hash_id) {
            last_idx = idx;
            return(lfi_hash_records[last_idx].lhr_hash_name);
        }
    }

    return("*ERROR*");
}


int
hash_name_to_hash_id(const char *hash_name)
{
    static int last_idx = -1;  /* a cache of our last lookup */
    int idx = -1;

    if (!hash_name) {
        return(-1);
    }

    if (last_idx != -1 && 
        !strcasecmp(lfi_hash_records[last_idx].lhr_hash_name, hash_name)) {
        return(lfi_hash_records[last_idx].lhr_hash_id);
    }

    for (idx = 0; 
         idx < (int) (sizeof(lfi_hash_records)/sizeof(lfi_hash_records[0]));
         idx++) {

        if (!strcasecmp(lfi_hash_records[idx].lhr_hash_name, hash_name)) {
            last_idx = idx;
            return(lfi_hash_records[last_idx].lhr_hash_id);
        }
    }

    return(-1);
}

const lfi_hash_record *
hash_record(int hash_id)
{
    int idx = -1;
    static int last_idx = -1;  /* a cache of our last lookup */

    if (hash_id < 0) {
        return(NULL);
    }

    if (last_idx != -1 && 
        lfi_hash_records[last_idx].lhr_hash_id == hash_id) {
        return(&lfi_hash_records[last_idx]);
    }

    for (idx = 0; 
         idx < (int) (sizeof(lfi_hash_records)/sizeof(lfi_hash_records[0]));
         idx++) {

        if (lfi_hash_records[idx].lhr_hash_id == hash_id) {
            last_idx = idx;
            return(&lfi_hash_records[last_idx]);
        }
    }

    return(NULL);
}

/*
 * This function hashes either a file on disk, or the passed in buffer.
 * Exactly one of the filename or the input buffer must be non-NULL, and
 * the other NULL.
 *
 * For regular files, the filename will be set, and buffer NULL, for symlinks
 * the filename will be NULL and the buffer the target of the symlink.
 *
 * Return 0 on success, 2 for file access errrors, 1 for all other errors.
 *
 */

static int
do_childhash(const char *filename, const char *in_buffer,
             int hash_id,
             char *ret_buf, int ret_buf_size,
             char *ret_error_buf, int ret_error_buf_size)
{
    int fnl = 0;
    ssize_t rr = 0;
    int pid = -1, status = -1;
    int stdout_pipe[2] = {-1, -1};
    int stderr_pipe[2] = {-1, -1};
    int stdin_pipe[2] = {-1, -1};
    const char *argv[LFI_MAX_HASH_PROG_ARGS + 1];
    int err = 0;
    const lfi_hash_record *hr = NULL;
    int arg = 0;
    ssize_t min_length = 0, start_offset = 0;

    if (!ret_error_buf) {
        return(1);
    }

    /* Make sure we never return junk unless we have to */
    if (ret_buf && ret_buf_size >= 1) {
        ret_buf[0] = '\0';
    }
    
    if ( (!filename && !in_buffer) ||  (filename && in_buffer) ) {
        snprintf(ret_error_buf, ret_error_buf_size,
                 "** ERROR: must specify exactly one of filename and "
                 "buffer.\n");
        return (1);
    }

    if (filename && (fnl = strlen(filename)) > FILENAME_MAX) {
        snprintf(ret_error_buf, ret_error_buf_size, 
                 "** ERROR: filename too long: %s\n", filename);
        return (1);
    }

    /* 
     * So we can tell if the error is access or not.  Yes, there is a
     * race here, but if so, it will just mean exiting more severely in
     * the non-root case, not doing something wrong.
     */
    if (filename && access(filename, R_OK)) {
        snprintf(ret_error_buf, ret_error_buf_size, 
                 "** ERROR: cannot read %s: %s", filename,
                 strerror(errno));
        return (2);
    }

    /* Set up our argv from the hash record */
    hr = hash_record(hash_id);
    if (!hr) {
        snprintf(ret_error_buf, ret_error_buf_size, 
                 "** ERROR: hash record: %d\n", hash_id);
        return (1);
    }

    /* Just so we get a better error message while debugging */
    if (access(hr->lhr_prog_binary_path, X_OK) != 0) {
        snprintf(ret_error_buf, ret_error_buf_size, 
                 "** ERROR: cannot execute %s: %s", hr->lhr_prog_binary_path,
                 strerror(errno));
        return (1);
    }

    argv[0] = NULL;
    if (filename) {
        for (arg = 0; 
             arg < LFI_MAX_HASH_PROG_ARGS && hr->lhr_prog_args_filename[arg]; 
             arg++) {
            if (strlen(hr->lhr_prog_args_filename[arg]) > 
                LFI_MAX_HASH_ARG_LENGTH) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: hash record arg: %d %d\n",
                         hash_id, arg);
                return (1);
            }
            if (!strcmp(hr->lhr_prog_args_filename[arg], "%p")) {
                argv[arg] = filename;
            }
            else {
                argv[arg] = hr->lhr_prog_args_filename[arg];
            }
        }
        argv[arg] = NULL;
    }
    else {
        /* The stdin / symlink (non-filename) case */
        for (arg = 0; 
             arg < LFI_MAX_HASH_PROG_ARGS && hr->lhr_prog_args_stdin[arg]; 
             arg++) {
            if (strlen(hr->lhr_prog_args_stdin[arg]) > 
                LFI_MAX_HASH_ARG_LENGTH) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: hash record arg: %d %d\n",
                         hash_id, arg);
                return (1);
            }
            if (!strcmp(hr->lhr_prog_args_stdin[arg], "%p")) {
                argv[arg] = hr->lhr_hash_parse_stdin_filename;
            }
            else {
                argv[arg] = hr->lhr_prog_args_stdin[arg];
            }
        }
        argv[arg] = NULL;
    }

    err = pipe(stdout_pipe);
    if (err < 0) {
        snprintf(ret_error_buf, ret_error_buf_size, "** ERROR: pipe: %s\n",
                 strerror(errno));
        return (1);
    }
    err = pipe(stderr_pipe);
    if (err < 0) {
        snprintf(ret_error_buf, ret_error_buf_size, "** ERROR: pipe: %s\n",
                 strerror(errno));
        return (1);
    }

    if (!filename) {
        err = pipe(stdin_pipe);
        if (err < 0) {
            snprintf(ret_error_buf, ret_error_buf_size,
                     "** ERROR: pipe: %s\n", strerror(errno));
            return (1);
        }
        signal(SIGPIPE, SIG_IGN);
    }
    
    pid = fork();
    if (pid == -1)
        return 1;
    if (pid == 0) {
        if (!filename) {
            dup2(stdin_pipe[0], STDIN_FILENO);
            close(stdin_pipe[1]);
        }
        dup2(stdout_pipe[1], STDOUT_FILENO);
        dup2(stderr_pipe[1], STDERR_FILENO);
        close(stdout_pipe[0]);
        close(stderr_pipe[0]);

        execve(hr->lhr_prog_binary_path, (char **) argv, environ);

        exit(127);
    }

    if (in_buffer) {
        close(stdin_pipe[0]);
        stdin_pipe[0] = -1;
        rr = write(stdin_pipe[1], in_buffer, strlen(in_buffer));
        if ((size_t) rr != strlen(in_buffer)) {
            snprintf(ret_error_buf, ret_error_buf_size, 
                     "** ERROR: write rr = %ld, not %ld\n", (long) rr, 
                     (long) strlen(in_buffer)
                     );
            status = -1;
        }
        close(stdin_pipe[1]);
        stdin_pipe[1] = -1;
    }
    close(stdout_pipe[1]);
    stdout_pipe[1] = -1;
    close(stderr_pipe[1]);
    stderr_pipe[1] = -1;

    do {
        if (waitpid(pid, &status, 0) == -1) {
            if (errno != EINTR) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: waitpid: %s", strerror(errno));
                return(1);
            }
        } else {
            if (WIFEXITED(status) == 0 || 
                WEXITSTATUS(status) != 0) {
                snprintf(ret_error_buf, ret_error_buf_size,
                         "** ERROR: could not hash '%s'\n",
                         filename ? filename : "[stdin]");
                return(1);
            }


            /* 
             * Now parse the raw output to extract just the hash.
             * We'll put this into ret_buf.
             */
            ret_buf[0] = '\0';
            rr = read(stdout_pipe[0], ret_buf, ret_buf_size);
            if (rr < 0) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: read: rr = %ld, %s", (long) rr,
                         strerror(errno));
                return(1);
            }
            while (rr > 0 && ret_buf[rr - 1] == '\n') {
                ret_buf[rr - 1] = '\0';
                rr--;
            }

            if (filename) {
                min_length = hr->lhr_hash_length + 
                    hr->lhr_hash_parse_filename_extra_min + 
                    (hr->lhr_hash_parse_filename_has_path ? fnl : 0);

                start_offset = (hr->lhr_hash_parse_filename_use_start >= 0 ?
                                hr->lhr_hash_parse_filename_use_start :
                                rr + hr->lhr_hash_parse_filename_use_start);
            }
            else {
                min_length = hr->lhr_hash_length + 
                    hr->lhr_hash_parse_stdin_extra_min + 
                    (hr->lhr_hash_parse_stdin_has_path ? fnl : 0);

                start_offset = (hr->lhr_hash_parse_stdin_use_start >= 0 ?
                                hr->lhr_hash_parse_stdin_use_start :
                                rr + hr->lhr_hash_parse_stdin_use_start);
            }
            
            if (rr < min_length || rr < start_offset) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: rr = %ld, not at least "
                         "%ld and %ld: %s\n",
                         (long) rr, (long) min_length,
                         (long) start_offset, ret_buf);
                return(1);
            }
            if (start_offset > 0) {
                memmove(ret_buf, 
                        ret_buf + start_offset,
                        hr->lhr_hash_length);
            }
            ret_buf[hr->lhr_hash_length] = '\0';

            if (strlen(ret_buf) != hr->lhr_hash_length) {
                snprintf(ret_error_buf, ret_error_buf_size, 
                         "** ERROR: bad hash: %s",
                         ret_buf);
                return(1);
            }

            /* XXX should look at stderr_pipe[0] */

            break;
        }
    } while(1);

    if (stdin_pipe[0] != -1) {
        close(stdin_pipe[0]);
    }
    if (stdin_pipe[1] != -1) {
        close(stdin_pipe[1]);
    }
    if (stdout_pipe[0] != -1) {
        close(stdout_pipe[0]);
    }
    if (stderr_pipe[0] != -1) {
        close(stderr_pipe[0]);
    }

    return(0);
}

static int
stat_mode_to_file_type(int stat_mode)
{

    if (S_ISFIFO(stat_mode)) {
        return(FIELD_MATCH_TYPE_FIFO);
    }
    else if (S_ISCHR(stat_mode)) {
        return(FIELD_MATCH_TYPE_CHAR_DEV);
    }
    else if (S_ISDIR(stat_mode)) {
        return(FIELD_MATCH_TYPE_DIR);
    }
    else if (S_ISBLK(stat_mode)) {
        return(FIELD_MATCH_TYPE_BLOCK_DEV);
    }
    else if (S_ISREG(stat_mode)) {
        return(FIELD_MATCH_TYPE_FILE);
    }
    else if (S_ISLNK(stat_mode)) {
        return(FIELD_MATCH_TYPE_SYMLINK);
    }
#ifdef S_ISSOCK 
    else if (S_ISSOCK(stat_mode)) {
        return(FIELD_MATCH_TYPE_SOCKET);
    }
#endif

    return(0);
}



int
stat_string_create(const char *filename, const char *link_filename, 
                   const struct stat *ss, 
                   int hash_id,
                   int exclude_mask,
                   int ret_str_len, char *ret_str)
{
    char efb[2 * FILENAME_MAX];
    int fnsl = 0xfffffff;
    char hrb[4096];
    char erb[4096];
    int err = 0;
    char *rso = NULL;
    int rc = -1;

    /* Make sure we never return junk unless we have to */
    if (ret_str && ret_str_len >= 1) {
        ret_str[0] = '\0';
    }

    if (!filename | !ss || !ret_str || (fnsl = strlen(filename)) == 0 || 
        ret_str_len < fnsl + MIN_STAT_BUFFER_LEN) { 
        return (1);
    }

#if OPT_ESCAPE_FILENAME == 1
    escape_filename(filename, 1, sizeof(efb), efb);
#endif

    hrb[0] = '-';
    hrb[1] = '\0';
    if (S_ISREG(ss->st_mode)) {
        err = do_childhash(filename, NULL, hash_id, hrb, sizeof(hrb),
                            erb, sizeof(erb));
    } else if (S_ISLNK(ss->st_mode)) {
        err = do_childhash(NULL, link_filename, hash_id, hrb, sizeof(hrb),
                           erb, sizeof(erb));
    }

    /* 
     * Hash failed, likely due to perms.  Still emit a line, but with a
     * constructed error string.  We don't fail out in the non-root bad
     * permissions case.
     */
    if (err != 0) {
        if (has_verbose && erb[0]) {
            fprintf(stderr, "ERROR: %s\n", erb);
        }
        snprintf(hrb, sizeof(hrb), "-#ERROR#-");
        if (err == 2 && uid != 0) {
            err = 0;
        }
    }

    /* Output fields, one at a time so we can ignore some */
    rso = ret_str;

    /* Id */
    rc = snprintf(rso, ret_str_len - (ret_str - rso), "%s", LFI_ID);
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* dev: raw device number not interesting to most callers */
    /* ino: inode number not interesting to most callers */

    /* File type */
    rc = snprintf(rso, ret_str_len - (ret_str - rso), " %#o", 
                  (unsigned int) ((ss->st_mode & S_IFMT) >> S_FMT_SHIFT));
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* File perms (mode) */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_MODE)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %#o", 
                      (unsigned int) (ss->st_mode & ~S_IFMT));
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* nlink: number of hard links */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_NLINK)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %d", 
                      (unsigned int) ss->st_nlink);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* UID */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_UID)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %d", 
                      (unsigned int) ss->st_uid);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* GID */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_GID)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %d", 
                      (unsigned int) ss->st_gid);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* RDEV: only for block or character special files */
    if ( ( S_ISBLK(ss->st_mode) || S_ISCHR(ss->st_mode) ) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %lld", 
                      (long long) ss->st_rdev);
    }
    else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* Size: only for regular or symlink file types */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_SIZE)) &&
        ( S_ISREG(ss->st_mode) || S_ISLNK(ss->st_mode) ) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %lu", 
                      (unsigned long) ss->st_size);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* Access time: last access time not interesting to most callers */
    if (0) {  /* XXX: if (!(exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_ATIME)))*/
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %ld", 
                      (long) ss->st_atime);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* Mod time */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_MTIME)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %ld", 
                      (long) ss->st_mtime);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* Change time */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_CTIME)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %ld", 
                      (long) ss->st_mtime);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    /* Hash type and Hash */
    if (! (exclude_mask & FIELD_LFI_FLAG(FIELD_LFI_HASH)) ) {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " %s %s", 
                      (hrb[0] == '-') ? "-" : 
                      hash_id_to_hash_name(hash_id), hrb);
    } else {
        rc = snprintf(rso, ret_str_len - (ret_str - rso), " - -");
    }
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    
    /* Filename (possibly escaped) */
    rc = snprintf(rso, ret_str_len - (ret_str - rso), " %s", efb);
    rso += rc;
    if (ret_str - rso > ret_str_len) {
        return(1);
    }

    return (err);
}

int
stat_string_extract_fields(const char *stat_str, 
                           char **ret_id, char **ret_type,
                           char **ret_mode, char **ret_nlink,
                           char **ret_uid, char **ret_gid,
                           char **ret_rdev, char **ret_size,
                           char **ret_atime, char **ret_mtime,
                           char **ret_ctime, char **ret_hash_type,
                           char **ret_hash, char **ret_raw_filename,

                           int *ret_hash_id,
                           char **ret_filename)
{
    int err = 0;
    int cs = -1, ablen = 0;
    const char *cl = NULL, *prev_cl = NULL;
    char uefn[FILENAME_MAX + 1];
    char ab[FILENAME_MAX * 4 + 1];

    if (ret_id) { *ret_id =  NULL; }
    if (ret_type) { *ret_type =  NULL; }
    if (ret_mode) { *ret_mode =  NULL; }
    if (ret_nlink) { *ret_nlink =  NULL; }
    if (ret_uid) { *ret_uid =  NULL; }
    if (ret_gid) { *ret_gid =  NULL; }
    if (ret_rdev) { *ret_rdev =  NULL; }
    if (ret_size) { *ret_size =  NULL; }
    if (ret_atime) { *ret_atime =  NULL; }
    if (ret_mtime) { *ret_mtime =  NULL; }
    if (ret_ctime) { *ret_ctime =  NULL; }
    if (ret_hash_type) { *ret_hash_type =  NULL; }
    if (ret_hash) { *ret_hash =  NULL; }
    if (ret_raw_filename) { *ret_raw_filename =  NULL; }

    if (ret_hash_id) {
        *ret_hash_id =  -1;
    }
    if (ret_filename) { *ret_filename =  NULL; }

    if (!stat_str || !*stat_str) {
        return(1);
    }

    /* Make sure we match the version */
    if (strncmp(stat_str, LFI_ID, strlen(LFI_ID))) {
        return(1);
    }

    /* Now count in by words, until we get to the last (filename) */
    prev_cl = stat_str;
    for (cs = 0; prev_cl && *prev_cl && cs < FIELD_LFI_LAST; cs++) {
        if (cs + 1 != FIELD_LFI_FILENAME) {
            cl = strchr(prev_cl, ' ');
        }
        else {
            cl = NULL;
        }

        if (!cl) {
            ablen = min(sizeof(ab) - 1, strlen(prev_cl));
        }
        else {
            ablen = min((int) sizeof(ab) - 1, cl - prev_cl);

            /* Move past space */
            cl++;
        }

        strncpy(ab, prev_cl, ablen);
        ab[ablen] = '\0';

        switch (cs + 1) {
        case FIELD_LFI_ID:
            if (ret_id ) { *ret_id =  strdup(ab); }
            break;

        case FIELD_LFI_TYPE:
            if (ret_type ) { *ret_type =  strdup(ab); }
            break;

        case FIELD_LFI_MODE:
            if (ret_mode ) { *ret_mode =  strdup(ab); }
            break;

        case FIELD_LFI_NLINK:
            if (ret_nlink ) { *ret_nlink =  strdup(ab); }
            break;

        case FIELD_LFI_UID:
            if (ret_uid ) { *ret_uid =  strdup(ab); }
            break;

        case FIELD_LFI_GID:
            if (ret_gid ) { *ret_gid =  strdup(ab); }
            break;

        case FIELD_LFI_RDEV:
            if (ret_rdev ) { *ret_rdev =  strdup(ab); }
            break;

        case FIELD_LFI_SIZE:
            if (ret_size ) { *ret_size =  strdup(ab); }
            break;

        case FIELD_LFI_ATIME:
            if (ret_atime ) { *ret_atime =  strdup(ab); }
            break;

        case FIELD_LFI_MTIME:
            if (ret_mtime ) { *ret_mtime =  strdup(ab); }
            break;

        case FIELD_LFI_CTIME:
            if (ret_ctime ) { *ret_ctime =  strdup(ab); }
            break;

        case FIELD_LFI_HASH_TYPE:
            if (ret_hash_type ) { *ret_hash_type =  strdup(ab); }
            if (ret_hash_id) { *ret_hash_id = hash_name_to_hash_id(ab); }
            break;

        case FIELD_LFI_HASH:
            if (ret_hash ) { *ret_hash =  strdup(ab); }
            break;

        case FIELD_LFI_FILENAME:
            err = unescape_filename(ab, sizeof(uefn), uefn);
            if (err) {
                return(1);
            }
            if (ret_filename) { *ret_filename = strdup(uefn); }
        }

        prev_cl = cl;
    }


    if (cs < FIELD_LFI_LAST) {
        return(1);
    }

    return(err);
}

int
stats_string_diff_mask(const char *line1, const char *line2)
{
    int err = 0;
    char *l1f[FIELD_LFI_LAST];
    char *l2f[FIELD_LFI_LAST];
    int lc = 0;
    int diff = 0;
    int diff_mask = 0;

    if (!line1 || !line2) {
        return(-1);
    }

    if (!strcmp(line1, line2)) {
        return(0);
    }

    memset(l1f, 0, sizeof(l1f));
    memset(l2f, 0, sizeof(l2f));

    /* 
     * We walk each line, stopping at spaces.  For each word we first see
     * if either == "-".  If it does, the compare trivially succeeds.  
     * Otherwise, we use strcmp, and continue on.
     */

    err = stat_string_extract_fields(line1,
                                     &(l1f[0]), &(l1f[1]), &(l1f[2]),
                                     &(l1f[3]), &(l1f[4]), &(l1f[5]),
                                     &(l1f[6]), &(l1f[7]), &(l1f[8]),
                                     &(l1f[9]), &(l1f[10]), &(l1f[11]),
                                     &(l1f[12]), &(l1f[13]),
                                     NULL, NULL);
    if (err) {
        return(-1);
    }

    err = stat_string_extract_fields(line2,
                                     &(l2f[0]), &(l2f[1]), &(l2f[2]),
                                     &(l2f[3]), &(l2f[4]), &(l2f[5]),
                                     &(l2f[6]), &(l2f[7]), &(l2f[8]),
                                     &(l2f[9]), &(l2f[10]), &(l2f[11]),
                                     &(l2f[12]), &(l2f[13]),
                                     NULL, NULL);
    if (err) {
        goto done;
    }

    diff = 0;
    for (lc = 0; lc < FIELD_LFI_LAST - 1; lc++) {
        if (!l1f[lc] || !l2f[lc]) {
            diff = 1;
            diff_mask |= FIELD_LFI_FLAG(lc + 1);
            continue;
        }
        if (!strcmp(l1f[lc], "-") || !strcmp(l2f[lc], "-") ) {
            continue;
        }
        if (!strcmp(l1f[lc], l2f[lc])) {
            continue;
        }
        diff = 1;
        diff_mask |= FIELD_LFI_FLAG(lc + 1);
    }

done:
    for (lc = 0; lc < FIELD_LFI_LAST; lc++) {
        free(l1f[lc]);
        free(l2f[lc]);
    }

    if (diff) {
        return(diff_mask);
    } else {
        return(0);
    }
}


int
main(int argc, char * const argv[])
{
    int c = -1;
    int i = -1;
    const char *filename = NULL;
    struct stat sb;
    char ss[1000];
    char lfn[FILENAME_MAX + 1] = {'\0', };
    int sscr = -1;
    int rlc = -1;
    int err = 0;
    int offd = -1;
    FILE *offh = NULL, *ops = NULL, *cffh = NULL;
    short parse_error = 0;
    short has_output_file = 0, has_check_file = 0, has_check_root_dir = 0;
    short seen_diff = 0, had_problem = 0;
    int ignore_masks[FIELD_MATCH_COUNT + 2] = {0, };
    int ignore_mask = 0;
    int ftype = 0;
    char *output_file = NULL, *check_file = NULL, *check_root_dir = NULL;
    char line_buf[FILENAME_MAX + 100];
    char *fn = NULL;
    int hash_id = -1;
    int hash_check_id = -1;
    int cfc = 0, gfc = 0, bad_lines = 0, lnum = 0, missing_count = 0;
    int im = 0;
    int diff_mask = 0;
    int ret_code = 0;

    uid = getuid();
    hash_id = hash_name_to_hash_id(HASH_DEFAULT_NAME);

    /* Parse command line options */
    while (!parse_error && 
           (c = getopt(argc, argv, "o:c:r:qva:tTnNgGuUmM")) != -1) {
        switch(c) {
        case 'o':
            has_output_file = 1;
            if (optarg) {
                output_file = strdup(optarg);
            }
            else {
                parse_error = 1;
            }
            break;

        case 'c':
            has_check_file = 1;
            if (optarg) {
                check_file = strdup(optarg);
            }
            else {
                parse_error = 1;
            }
            break;

        case 'r':
            has_check_root_dir = 1;
            if (optarg) {
                check_root_dir = strdup(optarg);
            }
            else {
                parse_error = 1;
            }
            break;

        case 'q':
            has_quiet = 1;
            break;

        case 'v':
            has_verbose++;
            break;

        case 'a':
            if (optarg) {
                hash_id = hash_name_to_hash_id(optarg);
                if (hash_id == -1) {
                    parse_error = 1;
                }
            }
            else {
                parse_error = 1;
            }
            break;


        /* All of the following are about ignoring certain fields */

        case 't':
            ignore_masks[FIELD_MATCH_TYPE_ALL] |= 
                ((FIELD_LFI_FLAG(FIELD_LFI_ATIME)) |
                 (FIELD_LFI_FLAG(FIELD_LFI_MTIME)) |
                 (FIELD_LFI_FLAG(FIELD_LFI_CTIME)));
            break;

        case 'T':
            /* Ignore times for anything but files */
            im = (FIELD_LFI_FLAG(FIELD_LFI_ATIME)) |
                (FIELD_LFI_FLAG(FIELD_LFI_MTIME)) |
                (FIELD_LFI_FLAG(FIELD_LFI_CTIME));
 
            ignore_masks[FIELD_MATCH_TYPE_FIFO] |= im;
            ignore_masks[FIELD_MATCH_TYPE_CHAR_DEV] |= im;
            ignore_masks[FIELD_MATCH_TYPE_DIR] |= im;
            ignore_masks[FIELD_MATCH_TYPE_BLOCK_DEV] |= im;
            ignore_masks[FIELD_MATCH_TYPE_SYMLINK] |= im;
            ignore_masks[FIELD_MATCH_TYPE_SOCKET] |= im;
            break;

        case 'n':
            ignore_masks[FIELD_MATCH_TYPE_ALL] |= 
                FIELD_LFI_FLAG(FIELD_LFI_NLINK);
            break;

        case 'N':
            /* Ignore link count for directories only */
            ignore_masks[FIELD_MATCH_TYPE_DIR] |= 
                FIELD_LFI_FLAG(FIELD_LFI_NLINK);
            break;

        case 'u':
            ignore_masks[FIELD_MATCH_TYPE_ALL] |= 
                FIELD_LFI_FLAG(FIELD_LFI_UID);
            break;

        case 'U':
            /* Ignore user for device files */
            ignore_masks[FIELD_MATCH_TYPE_CHAR_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_UID);
            ignore_masks[FIELD_MATCH_TYPE_BLOCK_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_UID);
            break;

        case 'g':
            ignore_masks[FIELD_MATCH_TYPE_ALL] |= 
                FIELD_LFI_FLAG(FIELD_LFI_GID);
            break;

        case 'G':
            /* Ignore group for device files */
            ignore_masks[FIELD_MATCH_TYPE_CHAR_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_GID);
            ignore_masks[FIELD_MATCH_TYPE_BLOCK_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_GID);
            break;

        case 'm':
            ignore_masks[FIELD_MATCH_TYPE_ALL] |= 
                FIELD_LFI_FLAG(FIELD_LFI_MODE);
            break;

        case 'M':
            /* Ignore mode for device files */
            ignore_masks[FIELD_MATCH_TYPE_CHAR_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_MODE);
            ignore_masks[FIELD_MATCH_TYPE_BLOCK_DEV] |= 
                FIELD_LFI_FLAG(FIELD_LFI_MODE);
            break;


        case '?':
            parse_error = 1;
            break;
        }
    }

    if (has_check_root_dir && !has_check_file) {
        parse_error = 1;
    }

    if (has_check_file && optind != argc) {
        parse_error = 1;
    }

    if (has_quiet && (has_output_file || !has_check_file)) {
        parse_error = 1;
    }

    if (has_quiet && has_verbose) {
        parse_error = 1;
    }

    if (!has_check_file && optind >= argc) {
        parse_error = 1;
    }

    if (parse_error) {
        fprintf(stderr, "Usage: \n"
          "  lfi [-o output_file] \n"
          "      [-q] [-v] [-t] [-T] [-g] [-G] [-u] [-U] [-n] [-N] [-m] [-M]\n"
          "      FILE [FILE]...\n");
        fprintf(stderr, 
          "  lfi -c check_file [-r check_root_dir] \n"
          "      [-o output_file] [-q] [-v] [-t] [-T] [-g] [-G] \n"
          "      [-u] [-U] [-n] [-N] [-m] [-M] \n");
        fprintf(stderr, "\n");
        fprintf(stderr, 
          "  lfi has two modes, generate and verify.  By default, \n"
          "  lfi will generate new file fingerprint output .  This \n"
          "  can be used to make a single output file encodes for each of \n"
          "  the specified input files the meta-information (file \n"
          "  size, mode, modification time, etc) as well as a hash of the \n"
          "  file contents.  This can later be used (with verify mode) to \n"
          "  verify that no files have changed.  This may be useful as part \n"
          "  of a backup/restore scheme, or as part of a system integrity \n"
          "  tool.\n"
          "  \n"
          "Generate mode is the default.  Verify mode is selected with '-c'.\n"
          "  \n"
          "Verify mode only:\n"
          "  -c CHECK_FILE : enables verify of previously generated lfi\n"
          "      output stored in CHECK_FILE.  Will exit with 0 on success.\n"
          "  -r CHECK_ROOT_DIR : change to this directory before compare\n"
          "  \n"
          "Verify and Generate modes:\n"
          "  -q : quiet operation\n"
          "  -v : verbose operation\n"
          "  -o OUTPUT_FILE : output results to OUTPUT_FILE\n"
          "  -a ALGORITHM : hash algorithm: md5 [default], sha1, sha256\n"
          "  -t : ignore times\n"
          "  -T : ignore times for everything except files\n"
          "  -g : ignore file group\n"
          "  -G : ignore file group for device files only\n"
          "  -u : ignore file user\n"
          "  -U : ignore file user for device files\n"
          "  -n : ignore number of links\n"
          "  -N : ignore number of links for directories\n"
          "  -m : ignore file mode\n"
          "  -M : ignore file mode for device files only\n");
        return(1);
    }

    ops = stderr;
    if (has_output_file) {
        offd = open(output_file, O_WRONLY | O_CREAT | O_TRUNC, 0644);
        if (offd < 0) {
            fprintf(stderr, "ERROR: open with '%s' for %s\n", strerror(errno),
                    output_file);
            return(1);
        }
        offh = fdopen(offd, "w");
        if (!offh) {
            fprintf(stderr, "ERROR: fdopen with '%s' for %s\n", 
                    strerror(errno), output_file);
            return(1);
        }

        ops = offh;
    }

    if (has_check_file) {
        if (!strcmp(check_file, "-")) { /* Use stdin */
            cffh = stdin;
        }
        else {
            cffh = fopen(check_file, "r");
            if (!cffh) {
                fprintf(stderr, "ERROR: open with '%s' for %s\n", 
                        strerror(errno), check_file);
                return(1);
            }
        }
    }

    if (has_check_root_dir) {
        err = chdir(check_root_dir);
        if (err < 0) {
            fprintf(stderr, "ERROR: chdir with '%s' for %s\n", 
                    strerror(errno), check_root_dir);
            return(1);
        }
    }

    /* The normal case, output a bunch of file info */
    if (!has_check_file) {
        for (i = optind; i < argc; i++) {
            filename = argv[i];

            err = lstat(filename, &sb);  /* lstat (not stat) for symlinks */
            if (err < 0) {

                if (errno == ENOENT) {
                    if (!has_quiet) {
                        fprintf(ops, "ERROR-MISSING: %s\n", filename);
                        fflush(ops);
                    }
                    had_problem = 1;
                    continue;
                }

                if (!has_quiet) {
                    fprintf(ops, "ERROR: lstat with '%s' for %s\n", 
                            strerror(errno), filename);
                    fflush(ops);
                }
                had_problem = 1;
                continue;
            }
            if (S_ISLNK(sb.st_mode)) {
                rlc = readlink(filename, lfn, sizeof(lfn) - 1);
                if (rlc > 0) {
                    lfn[rlc] ='\0';
                }
                else {
                    if (!has_quiet) {
                        fprintf(ops, "ERROR: readlink with '%s' for %s\n", 
                                strerror(errno), filename);
                        fflush(ops);
                    }
                    had_problem = 1;
                    continue;
                }
            }

            ss[0] = '\0';
            ftype = stat_mode_to_file_type(sb.st_mode);
            ignore_mask = (ignore_masks[ftype] == 0) ? 
                ignore_masks[FIELD_MATCH_TYPE_ALL] : 
                ignore_masks[ftype];
            sscr = stat_string_create(filename, lfn, &sb, hash_id, 
                                      ignore_mask,
                                      sizeof(ss), ss);
        
            if (sscr == 0) {
                if (has_output_file) {
                    write(offd, ss, strlen(ss));
                    write(offd, "\n", 1);
                }
                else {
                    fprintf(stdout, "%s\n", ss);
                    fflush(stdout);
                }
            }
            else {
                if (!has_quiet) {
                    fprintf(ops, "ERROR: hash gen for %s\n", filename);
                    fflush(ops);
                }
                had_problem = 1;
            }
        }
    }
    else { /* has a check file */

        bad_lines = 0;
        lnum = 0;
        cfc = 0;
        do {
            if (fn) {
                free(fn);
                fn = NULL;
            }

            if (!fgets(line_buf, sizeof(line_buf), cffh)) {
                if (!feof(cffh)) {
                    if (!has_quiet) {
                        fprintf(ops, "ERROR: reading from %s\n", check_file);
                        fflush(ops);
                    }
                    had_problem = 1;
                }
                break;
            }

            lnum++;

            /* Ignore this line, it does not look close to right */
            if (line_buf[0] == '#' || strncmp(line_buf, LFI_PREFIX, 
                                              strlen(LFI_PREFIX))) {
                continue;
            }

            cfc++;

            if (line_buf[strlen(line_buf) - 1] == '\n') {
                line_buf[strlen(line_buf) - 1] = '\0';
            }
            
            err = stat_string_extract_fields(line_buf, 
                                             NULL, NULL,
                                             NULL, NULL,
                                             NULL, NULL,
                                             NULL, NULL,
                                             NULL, NULL,
                                             NULL, NULL,
                                             NULL, NULL,
                                             &hash_check_id, &fn);
            if (err) {
                bad_lines++;
                if (!has_quiet) {
                    fprintf(ops, "ERROR: bad line at %s line %d\n", 
                            check_file, lnum);
                    fflush(ops);
                }
                continue;
            }

            /* Now generate the string, and compare */
            err = lstat(fn, &sb);  /* lstat (not stat) for symlinks */
            if (err < 0) {
                if (errno == ENOENT) {
                    if (!has_quiet) {
                        fprintf(ops, "FAILED-MISSING: %s\n", fn);
                        fflush(ops);
                    }
                    missing_count++;
                    continue;
                }

                if (!has_quiet) {
                    fprintf(ops, "ERROR: lstat with '%s' for %s\n", 
                            strerror(errno), fn);
                    fflush(ops);
                }
                had_problem = 1;
                continue;
            }
            if (S_ISLNK(sb.st_mode)) {
                rlc = readlink(fn, lfn, sizeof(lfn) - 1);
                if (rlc > 0) {
                    lfn[rlc] ='\0';
                }
                else {
                    if (!has_quiet) {
                        fprintf(ops, "ERROR: readlink with '%s' for %s\n", 
                                strerror(errno), fn);
                        fflush(ops);
                    }
                    had_problem = 1;
                    continue;
                }
            }

            ss[0] = '\0';
            ftype = stat_mode_to_file_type(sb.st_mode);
            ignore_mask = (ignore_masks[ftype] == 0) ? 
                ignore_masks[FIELD_MATCH_TYPE_ALL] : 
                ignore_masks[ftype];
            sscr = stat_string_create(fn, lfn, &sb, hash_check_id, ignore_mask,
                                      sizeof(ss), ss);
        
            if ((diff_mask = stats_string_diff_mask(line_buf, ss))) {
                seen_diff = 1;
                if (!has_quiet) {
                    if (diff_mask & FIELD_LFI_FLAG(FIELD_LFI_HASH)) {
                        fprintf(ops, "FAILED-HASH: %s\n", fn);
                    }
                    else {
                        fprintf(ops, "FAILED-STAT: %s\n", fn);
                    }
                    if (has_verbose >= 2) {
                        fprintf(ops, "   FAILED: check:   %s\n", line_buf);
                        fprintf(ops, "   FAILED: current: %s\n", ss);
                    }
                    fflush(ops);
                }
                continue;
            }
            else {
                if (has_verbose) {
                    fprintf(ops, "OK: %s\n", fn);
                    fflush(ops);
                }
            }

            gfc++;
        } while (!feof(cffh) && !ferror(cffh));

        if (ferror(cffh)) {
            if (!has_quiet) {
                fprintf(ops, "ERROR: processing %s\n", check_file);
                fflush(ops);
            }
            had_problem = 1;
        }


        if (has_verbose) {
            fprintf(ops, "\n");
            fflush(ops);
        }

        ret_code = 0;

        if (bad_lines != 0) {
            if (!has_quiet) {
                fprintf(ops, "FAILED-BAD-LINES: had %d bad lines\n", 
                        bad_lines);
            }
            ret_code |= 8;
        } 

        if (cfc - gfc != 0 || seen_diff) {
            if (missing_count) {
                ret_code |= 2;
            }
            if (missing_count != cfc - gfc) {
                ret_code |= 4;
            }

            if (!has_quiet) {
                if (missing_count == cfc - gfc) {
                    fprintf(ops, "FAILED-MISSING-ONLY: %d of %d files "
                            "missing\n",
                            cfc - gfc, cfc);
                }
                else {
                    fprintf(ops, "FAILED: %d of %d files (%d missing) did not "
                        "verify.\n",
                        cfc - gfc, cfc, missing_count);
                }
            }
        }
        else {
            if (!has_quiet) {
                if (had_problem) {
                    fprintf(ops, "FAILED: unknown problem with %d files.\n",
                            cfc);
                }
                else {
                    fprintf(ops, "VERIFIED: %d files match.\n", cfc);
                }
            }
        }

        fflush(ops);

        if (had_problem) {
            ret_code |= 1;
        }


        return(ret_code);
    }

    if (had_problem) {
        return(1);
    }

    return(0);
}
